import os
from pathlib import Path
from platform import python_version_tuple

import pytest
from hatchling.builders.plugin.interface import IncludedFile


@pytest.mark.slow
def test_build_with_real_sam(hatch):
    major, minor, _patch = python_version_tuple()
    assert major == "3"
    minor = min(int(minor), 9)
    build_conf = {"sam-params": ["--parameter-overrides", f"PythonVersion={major}.{minor}"]}
    builder = hatch(build_conf=build_conf)
    builder.build_standard(directory=builder.config.directory)

    dist = Path(f"{builder.root}/.aws-sam")

    assert (dist / "build.toml").is_file()
    assert (dist / "build" / "template.yaml").is_file()
    assert Path(
        f"{builder.root}/.aws-sam/build/MyLambda1Func/my_app/lambdas/lambda1/main.py"
    ).is_file()


def test_clean_method(hatch):
    builder = hatch()

    build_dir = Path(builder.config.directory)
    build_dir.mkdir(parents=True, exist_ok=True)
    app_file = build_dir / "data" / "app.py"
    app_file.parent.mkdir(parents=True, exist_ok=True)
    app_file.touch()

    builder.clean(directory=builder.config.directory, _versions=None)

    assert not build_dir.exists()


@pytest.mark.parametrize("template", ["bla", "template.yaml", "sam-template.yaml"])
def test_sam_template_config_option(hatch, template):
    config = {
        "project": {
            "name": "my-app",
            "version": "0.0.1",
        },
        "tool": {"hatch": {"build": {"targets": {"aws": {"template": template}}}}},
    }
    builder = hatch(config=config)

    assert builder.config.template == Path(builder.root) / template


@pytest.mark.slow
@pytest.mark.parametrize(
    "force_include, expected_files",
    (
        (
            {
                "src/my_app/common": "my_app/common",
                "src/my_app/utils/logger.py": "my_app/utils/logger.py",
            },
            [
                "my_app/lambdas/lambda1/main.py",
                "my_app/lambdas/lambda1/db.py",
                "my_app/common/config.py",
                "my_app/common/models.py",
                "my_app/utils/logger.py",
            ],
        ),
        (
            None,
            [
                "my_app/lambdas/lambda1/main.py",
                "my_app/lambdas/lambda1/db.py",
            ],
        ),
        (
            {"void/null/nothing": "void/null/nothing"},
            [
                "my_app/lambdas/lambda1/main.py",
                "my_app/lambdas/lambda1/db.py",
            ],
        ),
    ),
)
def test_build_lambda(hatch, force_include, expected_files, aws_lambda):
    builder = hatch(force_include=force_include, build=True)

    shared_files = []

    if force_include:
        for rel, dist in force_include.items():
            absolute_path = Path(builder.root) / rel
            if absolute_path.is_file():
                shared_files.append(
                    IncludedFile(distribution_path=dist, relative_path=rel, path=str(absolute_path))
                )
            for path in absolute_path.rglob("*"):
                if path.is_file():
                    shared_files.append(
                        IncludedFile(
                            distribution_path=dist, relative_path=rel, path=str(absolute_path)
                        )
                    )

    builder.build_lambda(aws_lambda=aws_lambda, shared_files=shared_files)
    dist_folder = Path(f"{builder.root}/.aws-sam/build/MyLambda1Func")

    expected = sorted([str(dist_folder / file) for file in expected_files])
    files_in_dist = [str(path) for path in dist_folder.rglob("*") if path.is_file()]

    assert sorted(files_in_dist) == expected


@pytest.mark.slow
@pytest.mark.parametrize(
    "deps", [{"dependencies": ["pytest"]}, {"optional_dependencies": {"mylambda1func": ["pytest"]}}]
)
def test_build_lambda_with_pip_requirements(hatch, aws_lambda, deps):
    builder = hatch(build=True, **deps)

    builder.build_lambda(aws_lambda=aws_lambda, shared_files=[])
    dist_folder = Path(f"{builder.root}/.aws-sam/build/MyLambda1Func")
    assert (dist_folder / "pytest").is_dir()


@pytest.mark.slow
def test_build_lambda_limited_src(hatch, limited_src_lambda):
    builder = hatch(build=True)

    builder.build_lambda(aws_lambda=limited_src_lambda, shared_files=[])

    dist_folder = Path(f"{builder.root}/.aws-sam/build/MyLambda3Func")
    files = os.listdir(dist_folder)
    assert sorted(files) == ["db.py", "main.py"]



from pathlib import Path
from shutil import copy
from subprocess import check_call
from sys import version_info
from tempfile import TemporaryDirectory
from typing import Any, Callable, Dict, Generator, List, Optional

# pylint: disable=redefined-outer-name
import pytest
import tomli_w

from hatch_aws.aws import AwsLambda
from hatch_aws.builder import AwsBuilder


@pytest.fixture
def temp_dir() -> Generator[Path, None, None]:
    with TemporaryDirectory() as directory:
        yield Path(directory).resolve()


@pytest.fixture
def asset() -> Callable:
    def _locate_asset(file: str) -> Path:
        return Path(__file__).parent.resolve() / "assets" / file

    return _locate_asset


def make_files(root: Path, files: List[str]) -> None:
    for file in files:
        path = root / file
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()


@pytest.fixture
def hatch(temp_dir, asset) -> Callable:
    default_config: Dict[str, Any] = {
        "project": {
            "name": "my-app",
            "version": "0.0.1",
        },
        "tool": {"hatch": {"build": {"targets": {"aws": {}}}}},
    }

    def _make_project(
        config: Optional[Dict] = None,
        dependencies: Optional[List] = None,
        optional_dependencies: Optional[Dict] = None,
        template: str = "sam-template.yml",
        build_conf: Optional[Dict] = None,
        force_include: Optional[Dict] = None,
        files: Optional[List[str]] = None,
        build: bool = False,
    ):
        if not config:
            config = default_config
        if dependencies:
            config["project"]["dependencies"] = dependencies
        if optional_dependencies:
            config["project"]["optional-dependencies"] = optional_dependencies

        if build_conf:
            config["tool"]["hatch"]["build"]["targets"]["aws"] = build_conf
        if force_include:
            config["tool"]["hatch"]["build"]["force-include"] = force_include

        builder = AwsBuilder(str(temp_dir), config=config)

        major, minor, *_ = version_info
        if (major, minor) < (3, 9):
            template = f"sam-template-{major}.{minor}.yml"

        copy(src=asset(template), dst=temp_dir / "template.yml")

        pyproject = temp_dir / "pyproject.toml"
        with open(pyproject, mode="wb") as file:
            tomli_w.dump(config, file)

        if not files:
            files = [
                "tests/test_app.py",
                "scripts/something.sh",
                "src/my_app/lambdas/lambda1/main.py",
                "src/my_app/lambdas/lambda1/db.py",
                "src/my_app/lambdas/lambda2/main.py",
                "src/my_app/lambdas/lambda3/main.py",
                "src/my_app/lambdas/lambda3/db.py",
                "src/my_app/common/config.py",
                "src/my_app/common/models.py",
                "src/my_app/utils/storage.py",
                "src/my_app/utils/logger.py",
            ]

        make_files(root=temp_dir, files=files)
        if build:
            check_call(
                [
                    "sam",
                    "build",
                    "--template",
                    temp_dir / "template.yml",
                    "--build-dir",
                    temp_dir / ".aws-sam/build",
                ]
            )

        return builder

    return _make_project


@pytest.fixture
def aws_lambda() -> AwsLambda:
    return AwsLambda(name="MyLambda1Func", path=Path("my_app/lambdas/lambda1"))


@pytest.fixture
def limited_src_lambda() -> AwsLambda:
    return AwsLambda(path=Path("."), name="MyLambda3Func")



import pytest

from hatch_aws.aws import Sam


def test_sam_init(asset):
    sam = Sam(sam_exec="sam", template=asset("sam-template.yml"))

    lambda1, lambda2, lambda3, *other = sam.lambdas

    assert lambda1.path.as_posix() == "my_app/lambdas/lambda1"
    assert lambda1.name == "MyLambda1Func"

    assert lambda2.path.as_posix() == "my_app/lambdas/lambda2"
    assert lambda2.name == "MyLambda2Func"

    assert lambda3.path.as_posix() == "."
    assert lambda3.name == "MyLambda3Func"

    assert not other


def test_unsupported_template(asset):
    with pytest.raises(AttributeError):
        Sam(sam_exec="sam", template=asset("sam-template-unsupported-handler.yml"))



from dataclasses import dataclass
from os import sep
from pathlib import Path
from subprocess import run  # nosec
from typing import Dict, List, Optional

import yaml


@dataclass
class AwsLambda:
    name: str
    path: Path


class Sam:  # pylint: disable=too-few-public-methods
    def __init__(self, sam_exec: str, template: Path):
        self.exec = sam_exec
        self.template_path = template
        self.template = self._parse_sam_template()
        self.lambdas = self._get_aws_lambdas()

    def _get_aws_lambdas(self) -> List[AwsLambda]:
        resources = self.template["Resources"]
        lambdas = {
            resource: {
                **self.template.get("Globals", {}).get("Function", {}),
                **resources[resource]["Properties"],
            }
            for resource, param in resources.items()
            if param["Type"] == "AWS::Serverless::Function"
        }
        return [
            AwsLambda(
                name=resource,
                path=Path(param["Handler"].replace(".", sep)).parent.parent,
            )
            for resource, param in lambdas.items()
            if param.get("Runtime", "").lower().startswith("python")
        ]

    def _parse_sam_template(self) -> Dict:
        yaml.SafeLoader.add_multi_constructor("!", lambda *args: None)
        return yaml.safe_load(self.template_path.read_text(encoding="utf-8"))

    def invoke_sam_build(self, build_dir: str, params: Optional[List[str]] = None):
        def_params = ["--template", str(self.template_path), "--build-dir", build_dir]
        if not params:
            params = []
        params.extend(def_params)

        return run(
            [self.exec, "build"] + params,
            text=True,
            encoding="utf-8",
            capture_output=True,
            check=False,
        )



from __future__ import annotations

import os
from pathlib import Path
from typing import List, Optional

from hatchling.builders.config import BuilderConfig

DEFAULT_TEMPLATE = "template.yml"


class AwsBuilderConfig(BuilderConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__directory = None
        self.__template = None
        self.__sam_exec = "sam"
        self.__use_sam = False
        self.__sam_params = None

    def default_exclude(self) -> list:
        return ["tests/"]

    @property
    def directory(self) -> str:
        if self.__directory is None:
            if "directory" in self.target_config:
                directory = self.target_config["directory"]
                if not isinstance(directory, str):
                    raise TypeError(
                        f"Field `tool.hatch.build.targets.{self.plugin_name}.directory` "
                        "must be a string"
                    )
            else:
                directory = self.build_config.get(
                    "directory", str(Path(self.root) / ".aws-sam/build")
                )
                if not isinstance(directory, str):
                    raise TypeError("Field `tool.hatch.build.directory` must be a string")
            self.__directory = self.normalize_build_directory(str(directory))

        return self.__directory

    @property
    def template(self) -> Path:
        if self.__template is None:
            template = Path(self.target_config.get("template", DEFAULT_TEMPLATE))
            if not template.is_absolute():
                template = self.root / template

            self.__template = template

        return self.__template

    @property
    def sam_exec(self) -> str:
        self.__sam_exec = self.target_config.get("sam_exec", self.__sam_exec)
        self.__sam_exec = os.getenv("HATCH_SAM_EXEC", self.__sam_exec)
        if not isinstance(self.__sam_exec, str):
            raise TypeError(
                f"Field `tool.hatch.build.targets.{self.plugin_name}.sam_exec` " "must be a string."
            )
        return self.__sam_exec

    @property
    def use_sam(self) -> bool:
        self.__use_sam = self.target_config.get("use-sam", self.__use_sam)
        if not isinstance(self.__use_sam, bool):
            raise TypeError(
                f"Field `tool.hatch.build.targets.{self.plugin_name}.use-sam` " "must be a boolean."
            )
        return self.__use_sam

    @property
    def sam_params(self) -> Optional[List[str]]:
        self.__sam_params = self.target_config.get("sam-params")
        if self.__sam_params is not None and not isinstance(self.__sam_params, list):
            raise TypeError(
                f"Field `tool.hatch.build.targets.{self.plugin_name}.sam-params` "
                "must be an array."
            )
        return self.__sam_params

    @property
    def only_packages(self) -> bool:
        return True

    def default_include(self):
        return list(self.force_include.keys())



import sys
from contextlib import suppress
from os import sep
from pathlib import Path
from shlex import quote
from shutil import copy, rmtree
from subprocess import PIPE, check_call  # nosec
from typing import Dict, List

from hatchling.builders.plugin.interface import BuilderInterface, IncludedFile
from hatchling.metadata.utils import normalize_project_name

from hatch_aws.aws import AwsLambda, Sam
from hatch_aws.config import AwsBuilderConfig


def matches_shared_file(path: Path, build_dir: Path, shared_files: List[IncludedFile]):
    dist_path = (build_dir / file.distribution_path for file in shared_files)
    return any(file in path.parents or file == path for file in dist_path)


class AwsBuilder(BuilderInterface):
    """
    Build AWS Lambda Functions source code
    """

    PLUGIN_NAME = "aws"

    def get_version_api(self) -> Dict:
        return {"standard": self.build_standard}

    def clean(self, directory: str, _versions):
        rmtree(directory)

    def build_lambda(self, aws_lambda: AwsLambda, shared_files: List[IncludedFile]) -> None:
        build_dir = Path(self.config.directory) / aws_lambda.name
        target = build_dir / aws_lambda.path
        dirs: List[Path] = []
        for path in build_dir.rglob("*"):
            if path.is_dir():
                dirs.append(path)
                continue
            if not any(
                (
                    target in path.parents,
                    matches_shared_file(build_dir=build_dir, path=path, shared_files=shared_files),
                )
            ):
                path.unlink(missing_ok=True)
        for folder in sorted(dirs, reverse=True):
            with suppress(OSError):
                folder.rmdir()
        deps = self.get_dependencies(module_name=normalize_project_name(aws_lambda.name))

        if not deps:
            return
        requirements_file = target / "requirements.txt"
        requirements_file.parent.mkdir(exist_ok=True, parents=True)
        requirements_file.write_text(data="\n".join(deps), encoding="utf-8")
        check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "-r",
                quote(str(requirements_file)),
                "-t",
                quote(str(build_dir)),
            ],
            stdout=PIPE,
            stderr=PIPE,
            shell=False,
        )

    def build_standard(self, directory: str, **_build_data) -> str:
        self.config: AwsBuilderConfig
        try:
            sam = Sam(sam_exec=self.config.sam_exec, template=self.config.template)
        except AttributeError:
            self.app.abort(
                "Unsupported type for a 'CodeUri' or 'Handler'. Only string is supported."
                "Functions !Sub, !Ref and others are not supported yet. "
            )
            raise
        self.app.display_waiting("Building lambda functions ...")
        result = sam.invoke_sam_build(
            build_dir=self.config.directory, params=self.config.sam_params
        )
        if result.returncode != 0:
            self.app.display_error(result.stderr)
            self.app.abort("SAM build failed!")

        if self.config.use_sam:
            return directory

        shared_files = list(self.recurse_included_files())
        for aws_lambda in sam.lambdas:
            self.app.display_info(f"{aws_lambda.name} ...", end=" ")
            self.build_lambda(aws_lambda=aws_lambda, shared_files=shared_files)
            self.app.display_success("success")

        if self.config.force_include:
            build_dir = Path(self.config.directory)
            for file in shared_files:
                if not "*" in file.distribution_path:
                    copy(src=file.path, dst=build_dir / file.distribution_path)
                    continue
                *glob, filename = file.distribution_path.rpartition("*")
                for path in build_dir.glob(pattern="".join(glob)):
                    if path.is_file():
                        continue
                    if filename.startswith(sep):
                        filename = filename[1:]
                    target = path / filename
                    target.parent.mkdir(exist_ok=True, parents=True)
                    if not target.exists():
                        copy(src=file.path, dst=target)
        self.app.display_success("Build successfull 🚀")
        return directory

    def get_dependencies(self, module_name: str) -> List[str]:
        return sorted(
            [
                *self.metadata.core.dependencies,
                *self.metadata.core.optional_dependencies.get(module_name, []),
            ]
        )

    @classmethod
    def get_config_class(cls):
        return AwsBuilderConfig
